/* ---------------------------------------------------------- 
%   (C)1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */
/* random number generator */

#include <klic/basic.h>  /* NRAND48, fatalf */

#ifdef NRAND48
#include <stdio.h>

#include <klic/struct.h>
#include <klic/g_methtab.h>
#include <klic/g_basic.h>
#include <klic/gcobject.h>

#define CLASS_NAME		random__numbers
#define GC_OBJ_TYPE		struct random_number_generator
#define GC_OBJ_SIZE(obj)  G_SIZE_IN_Q(GC_OBJ_TYPE)

#include <klic/gc_macro.h>

GC_OBJ_TYPE {
  struct consumer_object_method_table* method_table;
  long range;
  long max;
  unsigned short state[3];
};

GCDEF_UNIFY()
{
  G_STD_DECL;

 top:
  if( GC_TERM == NILATOM ){
    GC_TERMINATE;
  }else if( isref(GC_TERM) ){
    q temp = derefone(GC_TERM);
    if( isref(temp) && (temp == GC_TERM || derefone(temp) == GC_TERM) ){
      GC_RETURN_WITH_HOOK(GC_TERM);
    }else{
      GC_TERM = temp;
      goto top;
    }
  }else if( iscons(GC_TERM) ){
    long one_random;

    do{
      one_random = nrand48(GC_SELF->state);
    }while( one_random >= GC_SELF->max );

    derefone(car_of(GC_TERM)) = makeint(one_random % GC_SELF->range);
    GC_TERM = cdr_of(GC_TERM);
    goto top;
  }else{
    fatalf("random_numbers: does not support message %k", GC_TERM);
  }
}

GCDEF_PRINT()
{
  G_STD_DECL;
  fprintf(g_fp, "RANDOM@%X", (q*) GC_SELF - new_space_top());
  return 0;
}

GCDEF_GC()
{
  G_STD_DECL;
  q* newself = klic_alloc(sizeof(GC_OBJ_TYPE) / sizeof(q));

  BCOPY(GC_SELF, newself, sizeof(GC_OBJ_TYPE));
  return newself;
}

#define GCUSE_MY_UNIFY
#define GCUSE_MY_PRINT
#define GCUSE_MY_GC

#include <klic/gc_methtab.h>

GCDEF_NEW()
{
  GC_STD_DECL_FOR_NEW;
  GC_OBJ_TYPE* obj;
  long seed;
  long range;

  if( GC_ARGC < 1 || 2 < GC_ARGC ){
    debug_printf("Wrong number of arguments (%d)", GC_ARGC);
    GC_ERROR_IN_NEW("");
  }

  GC_DEREF_FOR_NEW(GC_ARGV[0]);
  if( !isint(GC_ARGV[0]) ) GC_ERROR_IN_NEW("First argument is not integer");
  range = intval(GC_ARGV[0]);
  if( range < 1 ){
    debug_printf("Invalid first argument (%d)", range);
    GC_ERROR_IN_NEW("");
  }

  if( GC_ARGC > 1 ){
    GC_DEREF_FOR_NEW(GC_ARGV[1]);
    if( !isint(GC_ARGV[1]) ) GC_ERROR_IN_NEW("Second argument is not integer");
    seed = intval(GC_ARGV[1]);
  }else{
    seed = 0;
  }

  GCSET_NEWOBJ_FOR_NEW(obj, GC_OBJ_SIZE(obj));

  obj->state[0] = seed >> (sizeof(seed)*4);
  obj->state[1] = seed >> (sizeof(seed)*2);
  obj->state[2] = seed >> (sizeof(seed)*0);
  obj->range = range;
  obj->max = ((~0LU << 1) >> 1) / range * range;
  GC_RETURN_FROM_NEW(GC_MAKE_HOOK_VAR((struct consumer_object*) obj));
}
#endif  /* NRAND48 */
